# -*- coding: utf-8 -*-
"""Qt list widgets."""

from weakref import WeakKeyDictionary, WeakValueDictionary

from objetto import POST, PRE
from objetto.changes import ListInsert, ListMove
from objetto.objects import ListObject
from objetto.utils.reraise_context import ReraiseContext
from objetto.utils.type_checking import assert_is_subclass
from Qt import QtCore, QtWidgets
from six.moves import xrange as x_range

from .._mixins import OQWidgetMixin
from .._models.list import ListModelHeader, OQListModel
from .._views.list import OQListView

__all__ = ["OQWidgetListDefaultHeader", "OQWidgetList"]


_MAXIMUM_SIZE = (1 << 24) - 1


class OQWidgetListDefaultHeader(ListModelHeader):
    """
    Default header for :class:`objettoqt.widgets.OQWidgetList`.

    Inherits from:
      - :class:`objettoqt.models.ListModelHeader`
    """

    def data(self, obj, row, role=QtCore.Qt.DisplayRole):
        """
        Returns `None` for the :attr:`QtCore.Qt.DisplayRole`.

        :param obj: List object.
        :type obj: objetto.objects.ListObject

        :param row: Row.
        :type row: int

        :param role: Role.
        :type role: QtCore.Qt.ItemDataRole

        :return: Data.
        :rtype: str or objetto.bases.BaseObject
        """
        if role == QtCore.Qt.DisplayRole:
            return
        return super(OQWidgetListDefaultHeader, self).data(obj, row, role=role)


class _OQWidgetListModel(OQListModel):
    def setObj(self, obj):
        error = "can't call 'setObj' on internal model, use the widget's method instead"
        raise RuntimeError(error)


class OQWidgetList(OQListView):
    """
    List of widgets.

    Observes actions sent from an instance of :class:`objetto.bases.ListObject`.

    Inherits from:
      - :class:`objettoqt.views.OQListView`

    :param parent: Parent.
    :type parent: QtCore.QObject or None

    :param editor_widget_type: Editor widget class for items.
    :type editor_widget_type: type[objettoqt.mixins.OQWidgetMixin]

    :param header: Header (or None to use \
:class:`objettoqt.widgets.OQWidgetListDefaultHeader`).
    :type header: objettoqt.models.AbstractListModelHeader or None

    :param mime_type: Mime type.
    :type mime_type: str or None

    :param scrollable: Whether should be scrollable or change size according to editors.
    :type scrollable: bool
    """

    OBase = ListObject
    """
    **read-only class attribute**

    Minimum `objetto` object base requirement.

    :type: type[objetto.objects.ListObject]
    """

    def __init__(
        self,
        parent=None,
        editor_widget_type=None,
        header=None,
        mime_type=None,
        *args,
        **kwargs,
    ):
        super(OQWidgetList, self).__init__(parent=parent, *args, **kwargs)

        # Check required editor widget type.
        with ReraiseContext(TypeError, "didn't provide 'editor_widget_type' parameter"):
            assert_is_subclass(editor_widget_type, OQWidgetMixin)

        # Default header.
        if header is None:
            headers = (OQWidgetListDefaultHeader(),)
        else:
            headers = (header,)

        # Internal attributes.
        self.__editor_widget_type = editor_widget_type
        self.__delegate = _WidgetListDelegate(parent=self)
        self.__model = _OQWidgetListModel(
            parent=self, headers=headers, mime_type=mime_type
        )
        self.__fit_to_contents = False
        self.__minimum_fit_size = 0
        self.__maximum_fit_size = _MAXIMUM_SIZE
        self.__update_layout_timer = QtCore.QTimer()

        # Update layout timer.
        self.__update_layout_timer.setSingleShot(True)
        try:
            self.__update_layout_timer.setTimerType(QtCore.Qt.CoarseTimer)
        except AttributeError:
            pass
        self.__update_layout_timer.timeout.connect(self.__update_layout)

        # Set item delegate and internal model.
        super(OQWidgetList, self).setItemDelegate(self.__delegate)
        super(OQWidgetList, self).setModel(self.__model)

    @QtCore.Slot()
    def __update_layout(self):

        # Fit to contents, need to calculate fixed size.
        if self.__fit_to_contents:

            # Prepare initial information.
            size = 0
            maximum_size = self.__maximum_fit_size
            flow = self.flow()
            spacing = self.spacing()

            # Add margin to initial size.
            margins = self.contentsMargins()
            if flow == QtWidgets.QListView.LeftToRight:
                size += margins.left() + margins.right()
            else:
                size += margins.top() + margins.bottom()

            # For every widget.
            for widget in self.editors() or ():

                # Add widget's size depending on the flow.
                widget_size_hint = widget.sizeHint()
                if flow == QtWidgets.QListView.LeftToRight:
                    size += widget_size_hint.width()
                else:
                    size += widget_size_hint.height()
                size += 2 * spacing

                # Exceeded maximum fit size, clamp and stop adding.
                if maximum_size is not None and size > maximum_size:
                    size = maximum_size
                    if flow == QtWidgets.QListView.LeftToRight:
                        super(OQWidgetList, self).setHorizontalScrollBarPolicy(
                            QtCore.Qt.ScrollBarAsNeeded
                        )
                    else:
                        super(OQWidgetList, self).setVerticalScrollBarPolicy(
                            QtCore.Qt.ScrollBarAsNeeded
                        )
                    break

            # Did not exceed maximum fit size.
            else:
                if flow == QtWidgets.QListView.LeftToRight:
                    super(OQWidgetList, self).setHorizontalScrollBarPolicy(
                        QtCore.Qt.ScrollBarAlwaysOff
                    )
                else:
                    super(OQWidgetList, self).setVerticalScrollBarPolicy(
                        QtCore.Qt.ScrollBarAlwaysOff
                    )
                minimum_size = self.__minimum_fit_size
                if size < minimum_size:
                    size = minimum_size

            # Set fixed dimension based on the flow.
            if flow == QtWidgets.QListView.LeftToRight:
                super(OQWidgetList, self).setFixedWidth(size)
            else:
                super(OQWidgetList, self).setFixedHeight(size)

        # Emit layout changed signal.
        self.__model.layoutChanged.emit()

    def __updateLayout__(self):

        # De-bounce.
        if self.__fit_to_contents:
            self.__update_layout_timer.start(10)
        else:
            self.__update_layout_timer.start(1)

    @QtCore.Slot()
    def __fixScrolling(self):

        # Custom maximum fit size hasn't been set.
        if self.__maximum_fit_size != _MAXIMUM_SIZE:
            return

        horizontal_bar = self.horizontalScrollBar()
        vertical_bar = self.verticalScrollBar()
        horizontal_bar.valueChanged.disconnect(self.__fixScrolling)
        vertical_bar.valueChanged.disconnect(self.__fixScrolling)
        flow = self.flow()
        try:
            if flow == QtWidgets.QListView.LeftToRight:
                horizontal_bar.setValue(horizontal_bar.minimum())
            else:
                vertical_bar.setValue(vertical_bar.minimum())
        finally:
            horizontal_bar.valueChanged.connect(self.__fixScrolling)
            vertical_bar.valueChanged.connect(self.__fixScrolling)

    def __onObjChanged__(self, obj, old_obj, phase):
        super(OQWidgetList, self).__onObjChanged__(obj, old_obj, phase)

        if phase is PRE:
            super(_OQWidgetListModel, self.__model).setObj(None)
        elif phase is POST and obj is not None:
            super(_OQWidgetListModel, self.__model).setObj(obj)
            for i, value in enumerate(obj):
                self.openPersistentEditor(self.__model.index(i))

    def __onActionReceived__(self, action, phase):
        super(OQWidgetList, self).__onActionReceived__(action, phase)

        if action.sender is self.__model.obj() and phase is POST:

            # Wait for the model to receive it first.
            self.__model.objToken().wait()

            # Open persistent editors when items are inserted.
            if isinstance(action.change, ListInsert):
                self.clearSelection()
                indexes = []
                for i in x_range(action.change.index, action.change.last_index + 1):
                    index = self.__model.index(i, 0, QtCore.QModelIndex())
                    self.openPersistentEditor(index)
                    indexes.append(index)

                if indexes:
                    if len(indexes) > 1:
                        self.select(
                            QtCore.QItemSelection(indexes[0], indexes[-1]),
                            QtCore.QItemSelectionModel.Select,
                            indexes[-1],
                        )
                    else:
                        self.select(
                            indexes[0], QtCore.QItemSelectionModel.Select, indexes[0]
                        )

            # Select moved items.
            elif isinstance(action.change, ListMove):
                first = self.__model.index(
                    action.change.post_index, 0, QtCore.QModelIndex()
                )
                if action.change.post_index != action.change.post_last_index:
                    last = self.__model.index(
                        action.change.post_last_index, 0, QtCore.QModelIndex()
                    )
                    selection = QtCore.QItemSelection(first, last)
                    current = last
                else:
                    selection = first
                    current = first
                self.select(
                    selection, QtCore.QItemSelectionModel.ClearAndSelect, current
                )

            # Update layout.
            self.__updateLayout__()

    def setItemDelegate(self, value):
        """
        Prevent setting item delegate.

        :param value: Item delegate.
        :type value: QtWidgets.QItemDelegate

        :raises RuntimeError: Always raised.
        """
        error = "can't set item delegate on '{}' object".format(type(self).__name__)
        raise RuntimeError(error)

    def setModel(self, model):
        """
        Prevent setting model.

        :param model: Model.
        :type model: QtCore.QAbstractItemModel

        :raises RuntimeError: Always raised.
        """
        error = "can't set model on '{}' object".format(type(self).__name__)
        raise RuntimeError(error)

    def setMinimumHeight(self, minimum_height):
        """
        Set minimum height.

        :param minimum_height: Minimum height.
        :type minimum_height: int
        """
        if self.__fit_to_contents and self.flow() != QtWidgets.QListView.LeftToRight:
            error = (
                "can't set minimum height when 'fitToContents' is turned on and flow "
                "is from top to bottom"
            )
            raise RuntimeError(error)
        super(OQWidgetList, self).setMinimumHeight(minimum_height)

    def setMaximumHeight(self, maximum_height):
        """
        Set maximum height.

        :param maximum_height: Maximum height.
        :type maximum_height: int
        """
        if self.__fit_to_contents and self.flow() != QtWidgets.QListView.LeftToRight:
            error = (
                "can't set maximum height when 'fitToContents' is turned on and flow "
                "is from top to bottom"
            )
            raise RuntimeError(error)
        super(OQWidgetList, self).setMaximumHeight(maximum_height)

    def setMinimumWidth(self, minimum_width):
        """
        Set minimum width.

        :param minimum_width: Minimum width.
        :type minimum_width: int
        """
        if self.__fit_to_contents and self.flow() == QtWidgets.QListView.LeftToRight:
            error = (
                "can't set minimum width when 'fitToContents' is turned on and flow "
                "is from left to right"
            )
            raise RuntimeError(error)
        super(OQWidgetList, self).setMinimumWidth(minimum_width)

    def setMaximumWidth(self, maximum_width):
        """
        Set maximum width.

        :param maximum_width: Maximum width.
        :type maximum_width: int
        """
        if self.__fit_to_contents and self.flow() == QtWidgets.QListView.LeftToRight:
            error = (
                "can't set maximum width when 'fitToContents' is turned on and flow "
                "is from left to right"
            )
            raise RuntimeError(error)
        super(OQWidgetList, self).setMaximumWidth(maximum_width)

    def setFixedSize(self, width_or_size, height=None):
        """
        Set fixed size.

        :param width_or_size: Width or size.
        :type width_or_size: int or QtCore.QSize

        :param height: Height or None.
        :type height: int or None
        """
        if height is None:
            width, height = width_or_size.width(), width_or_size.height()
        else:
            width, height = int(width_or_size), int(height)

        self.setMinimumWidth(width)
        self.setMaximumWidth(width)

        self.setMinimumHeight(height)
        self.setMaximumHeight(height)

    def setHorizontalScrollBarPolicy(self, policy):
        """
        Set horizontal scrollbar policy.

        :param policy: Policy.
        :type policy: QtCore.Qt.ScrollBarPolicy
        """
        if self.__fit_to_contents and self.flow() == QtWidgets.QListView.LeftToRight:
            error = (
                "can't set horizontal scrollbar policy when 'fitToContents' is turned "
                "on and flow is from left to right"
            )
            raise RuntimeError(error)
        super(OQWidgetList, self).setHorizontalScrollBarPolicy(policy)

    def setVerticalScrollBarPolicy(self, policy):
        """
        Set vertical scrollbar policy.

        :param policy: Policy.
        :type policy: QtCore.Qt.ScrollBarPolicy
        """
        if self.__fit_to_contents and self.flow() != QtWidgets.QListView.LeftToRight:
            error = (
                "can't set vertical scrollbar policy when 'fitToContents' is turned "
                "on and flow is from top to bottom"
            )
            raise RuntimeError(error)
        super(OQWidgetList, self).setVerticalScrollBarPolicy(policy)

    def setFlow(self, flow):
        """
        Set flow. Changing this will reset the scrollbars' policies and any fixed sizes.

        :param flow: Flow.
        :type flow: QtWidgets.QListView.Flow
        """
        fit_to_contents = self.fitToContents()
        self.setFitToContents(False)
        super(OQWidgetList, self).setFlow(flow)
        self.setFitToContents(fit_to_contents)

    def minimumFitSize(self):
        """
        Get minimum 'fit to contents' size.

        :return: Minimum 'fit to contents' size.
        :rtype: int
        """
        return self.__minimum_fit_size

    def setMinimumFitSize(self, minimum_size):
        """
        Set minimum 'fit to contents' size.

        :param minimum_size: Minimum 'fit to contents' size.
        :type minimum_size: int
        """
        self.__minimum_fit_size = int(minimum_size)
        if self.__minimum_fit_size > self.__maximum_fit_size:
            self.__maximum_fit_size = self.__minimum_fit_size
        self.__updateLayout__()

    def maximumFitSize(self):
        """
        Get maximum 'fit to contents' size.

        :return: Maximum 'fit to contents' size.
        :rtype: int
        """
        return self.__maximum_fit_size

    def setMaximumFitSize(self, maximum_size):
        """
        Set maximum 'fit to contents' size.

        :param maximum_size: Maximum 'fit to contents' size.
        :type maximum_size: int
        """
        self.__maximum_fit_size = int(maximum_size)
        if self.__maximum_fit_size < self.__minimum_fit_size:
            self.__minimum_fit_size = self.__maximum_fit_size
        self.__updateLayout__()

    def fitToContents(self):
        """
        Get whether this list is set to fit its contents or not.

        :return: True if fits its contents.
        :rtype: bool
        """
        return self.__fit_to_contents

    def setFitToContents(self, fit_to_contents=True):
        """
        Set whether this list is set to fit its contents or not.
        Changing this will reset the scrollbars' policies and any fixed sizes.

        :param fit_to_contents: True to fit its contents.
        :type fit_to_contents: bool
        """
        was_fit_to_contents = self.__fit_to_contents
        self.__fit_to_contents = bool(fit_to_contents)

        # No change.
        if was_fit_to_contents is self.__fit_to_contents:
            return

        # Going from 'fit to contents' to arbitrary size.
        if was_fit_to_contents:
            self.horizontalScrollBar().valueChanged.disconnect(self.__fixScrolling)
            self.verticalScrollBar().valueChanged.disconnect(self.__fixScrolling)
            super(OQWidgetList, self).setHorizontalScrollBarPolicy(
                QtCore.Qt.ScrollBarAsNeeded
            )
            super(OQWidgetList, self).setVerticalScrollBarPolicy(
                QtCore.Qt.ScrollBarAsNeeded
            )
            super(OQWidgetList, self).setMinimumHeight(0)
            super(OQWidgetList, self).setMaximumHeight(_MAXIMUM_SIZE)
            super(OQWidgetList, self).setMinimumWidth(0)
            super(OQWidgetList, self).setMaximumWidth(_MAXIMUM_SIZE)

        # Going from arbitrary size to 'fit to contents'.
        else:
            super(OQWidgetList, self).setHorizontalScrollBarPolicy(
                QtCore.Qt.ScrollBarAlwaysOff
            )
            super(OQWidgetList, self).setVerticalScrollBarPolicy(
                QtCore.Qt.ScrollBarAlwaysOff
            )
            self.horizontalScrollBar().valueChanged.connect(self.__fixScrolling)
            self.verticalScrollBar().valueChanged.connect(self.__fixScrolling)

            if self.flow() != QtWidgets.QListView.LeftToRight:
                super(OQWidgetList, self).setMinimumWidth(0)
                super(OQWidgetList, self).setMaximumWidth(_MAXIMUM_SIZE)
                super(OQWidgetList, self).setMinimumHeight(0)
                super(OQWidgetList, self).setMaximumHeight(0)
            else:
                super(OQWidgetList, self).setMinimumWidth(0)
                super(OQWidgetList, self).setMaximumWidth(0)
                super(OQWidgetList, self).setMinimumHeight(0)
                super(OQWidgetList, self).setMaximumHeight(_MAXIMUM_SIZE)

            self.__fixScrolling()

        # Update layout.
        self.__updateLayout__()

    def editors(self):
        """
        Get editor widgets.

        :return: Editor widgets.
        :rtype: tuple[objettoqt.mixins.OQWidgetMixin]
        """
        obj = self.obj()
        if not obj:
            return ()
        editors = []
        for value in list(obj):
            widget = self.itemDelegate().getEditor(value)
            if widget is None:
                return ()
            editors.append(widget)
        return tuple(editors)

    def resizeEvent(self, event):
        """
        Update layout on resize.

        :param event: Resize event.
        :type event: QtGui.QResizeEvent
        """
        super(OQWidgetList, self).resizeEvent(event)
        self.__updateLayout__()

    def editorWidgetType(self):
        """
        Get editor widget class.

        :return: Editor widget class.
        :rtype: type[objettoqt.mixins.OQWidgetMixin]
        """
        return self.__editor_widget_type

    def mimeType(self):
        """
        Get mime type.

        :return: Mime type.
        :rtype: str or None
        """
        return self.__model.mimeType()


class _WidgetListDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent):
        super(_WidgetListDelegate, self).__init__(parent=parent)
        self.__editors = WeakValueDictionary()
        self.__sizes = WeakKeyDictionary()
        self.__size_hints = WeakKeyDictionary()

    def createEditor(self, parent, option, index):
        widget = self.parent()
        if widget is not None:
            obj = widget.obj()
            if obj is not None:
                editor = widget.editorWidgetType()()
                editor.setParent(parent)
                value = obj[index.row()]
                editor.setObj(value)
                self.__editors[id(value)] = editor
                self.__sizes[editor] = editor.size()
                self.__size_hints[editor] = editor.sizeHint()
                return editor
        return QtWidgets.QLabel(parent=parent)

    def setEditorData(self, editor, index):
        widget = self.parent()
        if widget is not None:
            obj = widget.obj()
            if obj is not None:
                old_value = editor.obj()
                new_value = obj[index.row()]
                if old_value is not new_value:
                    self.__editors.pop(id(old_value), None)
                    self.__editors[id(new_value)] = editor
                    editor.setObj(new_value)
                    widget.__updateLayout__()

    def sizeHint(self, option, index):
        widget = self.parent()
        if widget is not None:
            obj = widget.obj()
            if obj is not None:
                row = index.row()
                value = obj[row]
                value_id = id(value)
                editor = self.__editors.get(value_id, None)
                if editor is not None:
                    size_hint = editor.sizeHint()
                    size = editor.size()
                    previous_size = self.__sizes[editor]
                    previous_size_hint = self.__size_hints[editor]
                    update_layout = False
                    if size != previous_size:
                        self.__sizes[editor] = size
                        update_layout = True
                    if size_hint != previous_size_hint:
                        self.__size_hints[editor] = size_hint
                        update_layout = True
                    if update_layout:
                        widget.__updateLayout__()
                    return size_hint
        return QtCore.QSize(0, 0)

    def getEditor(self, value):
        return self.__editors.get(id(value))



# -*- coding: utf-8 -*-
import pytest
from objetto.applications import Application
from objetto.objects import Object, attribute, list_cls
from Qt import QtWidgets

from objettoqt._models import ListModelHeader, OQListModel
from objettoqt.views import OQListView


def test_list_view():
    class Thing(Object):
        name = attribute(str, default="Foo")

    qt_app = QtWidgets.QApplication([])
    app = Application()
    initial = (Thing(app, name=str(i)) for i in range(1000))
    lst = list_cls(Thing)(app, initial)

    model = OQListModel(
        headers=(ListModelHeader(title="name"),), mime_type="application/thing_yaml"
    )
    model.setObj(lst)

    view = OQListView()
    view.setModel(model)

    view.show()
    qt_app.exec_()


if __name__ == "__main__":
    pytest.main([__file__, "-s", "-v"])



# -*- coding: utf-8 -*-
"""Widget."""

from Qt import QtWidgets

from .._mixins import OQWidgetMixin

__all__ = ["OQWidget"]


class OQWidget(OQWidgetMixin, QtWidgets.QWidget):
    """
    Mixed :class:`QtWidgets.QWidget` type.

    Observes actions sent from an instance of :class:`objetto.bases.BaseObject`.

    Inherits from:
      - :class:`objettoqt.mixins.OQWidgetMixin`
      - :class:`QtWidgets.QWidget`
    """



# -*- coding: utf-8 -*-
"""Widgets."""

from ._widgets import (
    OQHistoryWidget,
    OQHistoryWidgetDefaultHeader,
    OQWidget,
    OQWidgetList,
    OQWidgetListDefaultHeader,
)

__all__ = [
    "OQWidget",
    "OQWidgetListDefaultHeader",
    "OQWidgetList",
    "OQHistoryWidgetDefaultHeader",
    "OQHistoryWidget",
]



# -*- coding: utf-8 -*-
"""Mixed `Qt` object classes."""

from Qt import QtCore

from ._mixins import OQObjectMixin

__all__ = ["OQObject"]


class OQObject(OQObjectMixin, QtCore.QObject):
    """
    Mixed :class:`QtCore.QObject` type.

    Observes actions sent from an instance of :class:`objetto.bases.BaseObject`.

    Inherits from:
      - :class:`objettoqt.mixins.OQObjectMixin`
      - :class:`QtCore.QObject`
    """



# -*- coding: utf-8 -*-
import pytest
from objetto.applications import Application
from objetto.changes import Update
from objetto.constants import POST, PRE
from objetto.objects import Object, attribute, list_cls
from Qt import QtCore, QtWidgets

from objettoqt.mixins import OQWidgetMixin
from objettoqt.widgets import OQWidgetList


def test_widget_list():
    class Thing(Object):
        name = attribute(str, default="Foo")

    class ThingWidget(OQWidgetMixin, QtWidgets.QWidget):
        def __init__(self, **kwargs):
            super(ThingWidget, self).__init__(**kwargs)
            self.ui_label = QtWidgets.QLabel(parent=self)
            self.ui_layout = QtWidgets.QVBoxLayout()
            self.setLayout(self.ui_layout)

            self.ui_button_a = QtWidgets.QPushButton()
            self.ui_button_a.clicked.connect(self.contract)

            self.ui_button_b = QtWidgets.QPushButton()
            self.ui_button_b.clicked.connect(self.expand)

            self.ui_layout.addWidget(self.ui_button_a)
            self.ui_layout.addWidget(self.ui_label)
            self.ui_layout.addWidget(self.ui_button_b)

        @QtCore.Slot()
        def contract(self):
            self.ui_label.setContentsMargins(0, 0, 0, 0)

        @QtCore.Slot()
        def expand(self):
            self.ui_label.setContentsMargins(0, 100, 0, 100)

        def _onObjChanged(self, obj, old_obj, phase):
            if phase is PRE:
                self.ui_label.setText("")
            if phase is POST:
                if obj is not None:
                    self.ui_label.setText(obj.name)

        def _onActionReceived(self, action, phase):
            if action.sender is self.obj() and phase is POST:
                if isinstance(action.change, Update):
                    if "name" in action.change.new_values:
                        self.ui_label.setText(action.change.new_values["name"])

    qt_app = QtWidgets.QApplication([])
    app = Application()
    initial = (Thing(app, name=str(i)) for i in range(10))
    lst = list_cls(Thing)(app, initial)

    window = QtWidgets.QMainWindow()
    widget = QtWidgets.QWidget()
    window.setCentralWidget(widget)
    layout = QtWidgets.QHBoxLayout()
    widget.setLayout(layout)

    widget_list_a = OQWidgetList(
        editor_widget_type=ThingWidget, mime_type="application/thing_yaml"
    )
    widget_list_a.setObj(lst)

    widget_list_b = OQWidgetList(
        editor_widget_type=ThingWidget, mime_type="application/thing_yaml"
    )
    widget_list_b.setFitToContents(True)
    widget_list_b.setMinimumFitSize(32)
    widget_list_b.setObj(lst)

    widget_list_c = OQWidgetList(
        editor_widget_type=ThingWidget, mime_type="application/thing_yaml"
    )
    widget_list_c.setFitToContents(True)
    widget_list_c.setMinimumFitSize(32)
    widget_list_c.setMaximumFitSize(800)
    widget_list_c.setObj(lst)

    layout.addWidget(widget_list_a)
    layout.addWidget(widget_list_b)
    layout.addWidget(widget_list_c)

    window.show()

    qt_app.exec_()


if __name__ == "__main__":
    pytest.main([__file__, "-s", "-v"])



# -*- coding: utf-8 -*-
import pytest
from objetto.applications import Application
from objetto.objects import Object, attribute, list_cls
from Qt import QtWidgets
from six import integer_types, string_types

from objettoqt._models import ListModelHeader, OQListModel
from objettoqt.views import OQTreeListView


def test_list_tree_view():
    class Thing(Object):
        name = attribute(string_types, default="Foo")
        points = attribute(integer_types, default=1)

    qt_app = QtWidgets.QApplication([])
    app = Application()
    initial = (Thing(app, name=str(i), points=i + 3) for i in range(15))
    lst = list_cls(Thing)(app, initial)

    model = OQListModel(
        headers=(
            ListModelHeader(title="name"),
            ListModelHeader(title="points"),
        ),
        mime_type="application/thing_yaml",
    )
    model.setObj(lst)

    view = OQTreeListView()
    view.setModel(model)

    view.show()
    qt_app.exec_()


if __name__ == "__main__":
    pytest.main([__file__, "-s", "-v"])



# -*- coding: utf-8 -*-
import pytest
from objetto.applications import Application
from objetto.constants import POST
from objetto.objects import Object, attribute
from Qt import QtCore

from objettoqt.mixins import OQObjectMixin


def test_mixin():
    class Thing(Object):
        name = attribute(str, default="Foo")

    class DummyQObject(OQObjectMixin, QtCore.QObject):
        received = None
        changed = None

        def _onObjChanged(self, obj, old_obj, phase):
            self.changed = obj, old_obj, phase

        def _onActionReceived(self, action, phase):
            self.received = action, phase

    app = Application()
    thing = Thing(app)

    dummy = DummyQObject()
    assert dummy.objToken() is None
    dummy.setObj(thing)
    assert dummy.objToken() is not None

    assert dummy.changed is not None
    assert dummy.received is None
    thing.name = "Bar"
    assert dummy.received is not None
    assert dummy.received[-1] is POST

    dummy.setObj(None)
    assert dummy.changed == (None, thing, POST)
    assert dummy.objToken() is None
    dummy.received = None
    thing.name = "Foo"
    assert dummy.received is None
    assert dummy.changed is not None
    assert dummy.changed == (None, thing, POST)


if __name__ == "__main__":
    pytest.main([__file__])



# -*- coding: utf-8 -*-
"""Mixed `Qt` model classes."""

from ._models.list import AbstractListModelHeader, ListModelHeader, OQListModel

__all__ = [
    "OQListModel",
    "AbstractListModelHeader",
    "ListModelHeader",
]



# -*- coding: utf-8 -*-
"""History widget."""

from objetto import POST, PRE, data_attribute
from objetto.changes import Update
from objetto.history import HistoryObject
from Qt import QtCore, QtGui, QtWidgets
from six import string_types

from .._models import ListModelHeader, OQListModel
from .._views import OQTreeListView

__all__ = ["OQHistoryWidgetDefaultHeader", "OQHistoryWidget"]


class OQHistoryWidgetDefaultHeader(ListModelHeader):
    """
    Default header for :class:`objettoqt.widgets.OQHistoryWidget`.

    Inherits from:
      - :class:`objettoqt.models.ListModelHeader`
    """

    title = data_attribute(string_types, subtypes=True, default="name")
    """
    Title.

    :type: str
    """

    fallback = data_attribute(string_types, subtypes=True, default="---")
    """
    Fallback value.

    :type: str
    """

    def data(self, obj, row, role=QtCore.Qt.DisplayRole):
        """
        Dim/brighten text depending on the history's index.

        :param obj: History changes list.
        :type obj: objetto.objects.ListObject

        :param row: Row.
        :type row: int

        :param role: Role.
        :type role: QtCore.Qt.ItemDataRole

        :return: Data.
        :rtype: str or objetto.bases.BaseObject
        """

        # Dim/brighten text.
        history = obj._parent
        if isinstance(history, HistoryObject):
            if role == QtCore.Qt.ForegroundRole:
                if row > history.index:  # TODO: use system colors
                    return QtGui.QBrush(QtGui.QColor(128, 128, 138, 100))
                elif history.index == row:
                    return QtGui.QBrush(QtGui.QColor(255, 255, 255, 255))

        return super(OQHistoryWidgetDefaultHeader, self).data(obj, row, role=role)


class _OQHistoryWidgetModel(OQListModel):
    def setObj(self, obj):
        error = "can't call 'setObj' on internal model, use the widget's method instead"
        raise RuntimeError(error)


class OQHistoryWidget(OQTreeListView):
    """
    Mixed :class:`QtWidgets.QTreeView` type (for history objects).

    Observes actions sent from an instance of :class:`objetto.history.HistoryObject`.

    Inherits from:
      - :class:`objettoqt.views.OQTreeListView`

    :param parent: Parent.
    :type parent: QtCore.QObject or None

    :param headers: Headers (or None to use \
:class:`objettoqt.widgets.OQHistoryWidgetDefaultHeader`).
    :type headers: tuple[objettoqt.models.AbstractListModelHeader] or None

    :param extra_headers: Extra headers.
    :type extra_headers: collections.abc.Iterable[\
objettoqt.models.AbstractListModelHeader or str] or None

    :param use_tree_list_view: If True, will use a tree list view instead of a list.
    :type use_tree_list_view: bool or None
    """

    OBase = HistoryObject
    """
    **read-only class attribute**

    Minimum `objetto` object base requirement.

    :type: type[objetto.history.HistoryObject]
    """

    def __init__(self, parent=None, headers=None, mime_type=None, *args, **kwargs):
        super(OQHistoryWidget, self).__init__(parent=parent, *args, **kwargs)

        # Default headers.
        if headers is None:
            headers = (OQHistoryWidgetDefaultHeader(),)

        # Defaults.
        self.setWindowTitle("History")
        self.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.setAcceptDrops(False)
        self.setDragEnabled(mime_type is not None)
        self.setHeaderHidden(len(headers) <= 1)

        # Model.
        self.__model = _OQHistoryWidgetModel(
            parent=self, headers=headers, mime_type=mime_type
        )
        super(OQHistoryWidget, self).setModel(self.__model)

        # Signals.
        self.activated.connect(self.__onActivated)

    def __onObjChanged__(self, obj, old_obj, phase):
        super(OQHistoryWidget, self).__onObjChanged__(obj, old_obj, phase)

        if phase is PRE:
            super(_OQHistoryWidgetModel, self.__model).setObj(None)
        elif phase is POST and obj is not None:
            super(_OQHistoryWidgetModel, self.__model).setObj(obj.changes)

    def __onActionReceived__(self, action, phase):
        history = self.obj()
        if history is not None:
            if action.sender is history and phase is POST:
                change = action.change
                if isinstance(change, Update) and "index" in change.new_values:
                    old_index = change.old_values["index"]
                    new_index = change.new_values["index"]
                    first_index = min((old_index, new_index))
                    last_index = len(history.changes) - 1
                    self.__model.dataChanged.emit(
                        self.__model.index(first_index, 0, QtCore.QModelIndex()),
                        self.__model.index(last_index, 0, QtCore.QModelIndex()),
                    )

    @QtCore.Slot(QtCore.QModelIndex)
    def __onActivated(self, index):
        model = self.model()
        if model is None:
            return
        changes = model.obj()
        if changes is not None:
            history = changes._parent
            if history is not None and index.isValid():
                with history.app.write_context():
                    app = QtWidgets.QApplication.instance()
                    app.setOverrideCursor(QtCore.Qt.WaitCursor)
                    was_enabled = self.isEnabled()
                    self.setEnabled(False)
                    try:
                        history.set_index(index.row())
                    finally:
                        app.restoreOverrideCursor()
                        self.setEnabled(was_enabled)
                        if self.isVisible():
                            self.setFocus()

    def setModel(self, model):
        """
        Prevent setting model.

        :param model: Model.
        :type model: QtCore.QAbstractItemModel

        :raises RuntimeError: Always raised.
        """
        error = "can't set model on '{}' object".format(type(self).__name__)
        raise RuntimeError(error)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="objettoqt",
    version="0.1.4",
    author="Bruno Nicko",
    author_email="brunonicko@gmail.com",
    description="Utilities to use Objetto with Qt",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/brunonicko/objettoqt",
    packages=setuptools.find_packages(),
    install_requires=[
        "enum34; python_version < '3.4'",
        "pyrsistent",
        "qualname",
        "six",
        "slotted",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    python_requires=">=2.7",
)



# -*- coding: utf-8 -*-
import pytest
from objetto.applications import Application
from objetto.objects import list_cls
from Qt import QtCore

from objettoqt._models import OQListModel


def test_list_model():
    app = Application()
    lst = list_cls(int)(app, range(10))

    model = OQListModel()
    model.setObj(lst)

    assert model.rowCount() == len(lst) == 10
    for i, v in enumerate(lst):
        assert model.data(model.index(i), role=QtCore.Qt.UserRole) == v

    lst.extend(range(21, 11, -1))
    assert model.rowCount() == len(lst) == 20

    for i, v in enumerate(lst):
        assert model.data(model.index(i), role=QtCore.Qt.UserRole) == v


if __name__ == "__main__":
    pytest.main([__file__])
